"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const context_js_1 = __importDefault(require("./context.js"));
const constants_js_1 = require("./constants.js");
function parseGitDiff(diff, options) {
    const ctx = new context_js_1.default(diff, options);
    const files = parseFileChanges(ctx);
    return {
        type: 'GitDiff',
        files,
    };
}
exports.default = parseGitDiff;
function parseFileChanges(ctx) {
    const changedFiles = [];
    while (!ctx.isEof()) {
        const changed = parseFileChange(ctx);
        if (!changed) {
            break;
        }
        changedFiles.push(changed);
    }
    return changedFiles;
}
function parseFileChange(ctx) {
    if (!isComparisonInputLine(ctx.getCurLine())) {
        return;
    }
    ctx.nextLine();
    let isDeleted = false;
    let isNew = false;
    let isRename = false;
    let pathBefore = '';
    let pathAfter = '';
    while (!ctx.isEof()) {
        const extHeader = parseExtendedHeader(ctx);
        if (!extHeader) {
            break;
        }
        if (extHeader.type === constants_js_1.ExtendedHeader.Deleted)
            isDeleted = true;
        if (extHeader.type === constants_js_1.ExtendedHeader.NewFile)
            isNew = true;
        if (extHeader.type === constants_js_1.ExtendedHeader.RenameFrom) {
            isRename = true;
            pathBefore = extHeader.path;
        }
        if (extHeader.type === constants_js_1.ExtendedHeader.RenameTo) {
            isRename = true;
            pathAfter = extHeader.path;
        }
    }
    const changeMarkers = parseChangeMarkers(ctx);
    const chunks = parseChunks(ctx);
    if (isDeleted && changeMarkers) {
        return {
            type: constants_js_1.FileType.Deleted,
            chunks,
            path: changeMarkers.deleted,
        };
    }
    else if (isDeleted &&
        chunks.length &&
        chunks[0].type === 'BinaryFilesChunk') {
        return {
            type: constants_js_1.FileType.Deleted,
            chunks,
            path: chunks[0].pathBefore,
        };
    }
    else if (isNew && changeMarkers) {
        return {
            type: constants_js_1.FileType.Added,
            chunks,
            path: changeMarkers.added,
        };
    }
    else if (isNew && chunks.length && chunks[0].type === 'BinaryFilesChunk') {
        return {
            type: constants_js_1.FileType.Added,
            chunks,
            path: chunks[0].pathAfter,
        };
    }
    else if (isRename) {
        return {
            type: constants_js_1.FileType.Renamed,
            pathAfter,
            pathBefore,
            chunks,
        };
    }
    else if (changeMarkers) {
        return {
            type: constants_js_1.FileType.Changed,
            chunks,
            path: changeMarkers.added,
        };
    }
    else if (chunks.length &&
        chunks[0].type === 'BinaryFilesChunk' &&
        chunks[0].pathAfter) {
        return {
            type: constants_js_1.FileType.Changed,
            chunks,
            path: chunks[0].pathAfter,
        };
    }
    return;
}
function isComparisonInputLine(line) {
    return line.indexOf('diff') === 0;
}
function parseChunks(context) {
    const chunks = [];
    while (!context.isEof()) {
        const chunk = parseChunk(context);
        if (!chunk) {
            break;
        }
        chunks.push(chunk);
    }
    return chunks;
}
function parseChunk(context) {
    const chunkHeader = parseChunkHeader(context);
    if (!chunkHeader) {
        return;
    }
    if (chunkHeader.type === 'Normal') {
        const changes = parseChanges(context, chunkHeader.fromFileRange, chunkHeader.toFileRange);
        return {
            ...chunkHeader,
            type: 'Chunk',
            changes,
        };
    }
    else if (chunkHeader.type === 'Combined' &&
        chunkHeader.fromFileRangeA &&
        chunkHeader.fromFileRangeB) {
        const changes = parseChanges(context, chunkHeader.fromFileRangeA.start < chunkHeader.fromFileRangeB.start
            ? chunkHeader.fromFileRangeA
            : chunkHeader.fromFileRangeB, chunkHeader.toFileRange);
        return {
            ...chunkHeader,
            type: 'CombinedChunk',
            changes,
        };
    }
    else if (chunkHeader.type === 'BinaryFiles' &&
        chunkHeader.fileA &&
        chunkHeader.fileB) {
        return {
            type: 'BinaryFilesChunk',
            pathBefore: chunkHeader.fileA,
            pathAfter: chunkHeader.fileB,
        };
    }
}
function parseExtendedHeader(ctx) {
    const line = ctx.getCurLine();
    const type = constants_js_1.ExtendedHeaderValues.find((v) => line.startsWith(v));
    if (type) {
        ctx.nextLine();
    }
    if (type === constants_js_1.ExtendedHeader.RenameFrom || type === constants_js_1.ExtendedHeader.RenameTo) {
        return {
            type,
            path: line.slice(`${type} `.length),
        };
    }
    else if (type) {
        return {
            type,
        };
    }
    return null;
}
function parseChunkHeader(ctx) {
    const line = ctx.getCurLine();
    const normalChunkExec = /^@@\s\-(\d+),?(\d+)?\s\+(\d+),?(\d+)?\s@@\s?(.+)?/.exec(line);
    if (!normalChunkExec) {
        const combinedChunkExec = /^@@@\s\-(\d+),?(\d+)?\s\-(\d+),?(\d+)?\s\+(\d+),?(\d+)?\s@@@\s?(.+)?/.exec(line);
        if (!combinedChunkExec) {
            const binaryChunkExec = /^Binary\sfiles\s(.*)\sand\s(.*)\sdiffer$/.exec(line);
            if (binaryChunkExec) {
                const [all, fileA, fileB] = binaryChunkExec;
                ctx.nextLine();
                return {
                    type: 'BinaryFiles',
                    fileA: getFilePath(ctx, fileA, 'src'),
                    fileB: getFilePath(ctx, fileB, 'dst'),
                };
            }
            return null;
        }
        const [all, delStartA, delLinesA, delStartB, delLinesB, addStart, addLines, context,] = combinedChunkExec;
        ctx.nextLine();
        return {
            context,
            type: 'Combined',
            fromFileRangeA: getRange(delStartA, delLinesA),
            fromFileRangeB: getRange(delStartB, delLinesB),
            toFileRange: getRange(addStart, addLines),
        };
    }
    const [all, delStart, delLines, addStart, addLines, context] = normalChunkExec;
    ctx.nextLine();
    return {
        context,
        type: 'Normal',
        toFileRange: getRange(addStart, addLines),
        fromFileRange: getRange(delStart, delLines),
    };
}
function getRange(start, lines) {
    const startNum = parseInt(start, 10);
    return {
        start: startNum,
        lines: lines === undefined ? startNum : parseInt(lines, 10),
    };
}
function parseChangeMarkers(context) {
    const deleterMarker = parseMarker(context, '--- ');
    const deleted = deleterMarker
        ? getFilePath(context, deleterMarker, 'src')
        : deleterMarker;
    const addedMarker = parseMarker(context, '+++ ');
    const added = addedMarker
        ? getFilePath(context, addedMarker, 'dst')
        : addedMarker;
    return added && deleted ? { added, deleted } : null;
}
function parseMarker(context, marker) {
    const line = context.getCurLine();
    if (line?.startsWith(marker)) {
        context.nextLine();
        return line.replace(marker, '');
    }
    return null;
}
const CHAR_TYPE_MAP = {
    '+': constants_js_1.LineType.Added,
    '-': constants_js_1.LineType.Deleted,
    ' ': constants_js_1.LineType.Unchanged,
    '\\': constants_js_1.LineType.Message,
};
function parseChanges(ctx, rangeBefore, rangeAfter) {
    const changes = [];
    let lineBefore = rangeBefore.start;
    let lineAfter = rangeAfter.start;
    while (!ctx.isEof()) {
        const line = ctx.getCurLine();
        const type = getLineType(line);
        if (!type) {
            break;
        }
        ctx.nextLine();
        let change;
        const content = line.slice(1);
        switch (type) {
            case constants_js_1.LineType.Added: {
                change = {
                    type,
                    lineAfter: lineAfter++,
                    content,
                };
                break;
            }
            case constants_js_1.LineType.Deleted: {
                change = {
                    type,
                    lineBefore: lineBefore++,
                    content,
                };
                break;
            }
            case constants_js_1.LineType.Unchanged: {
                change = {
                    type,
                    lineBefore: lineBefore++,
                    lineAfter: lineAfter++,
                    content,
                };
                break;
            }
            case constants_js_1.LineType.Message: {
                change = {
                    type,
                    content: content.trim(),
                };
                break;
            }
        }
        changes.push(change);
    }
    return changes;
}
function getLineType(line) {
    return CHAR_TYPE_MAP[line[0]] || null;
}
function getFilePath(ctx, input, type) {
    if (ctx.options.noPrefix) {
        return input;
    }
    if (type === 'src')
        return input.replace(/^a\//, '');
    if (type === 'dst')
        return input.replace(/^b\//, '');
}
//# sourceMappingURL=parse-git-diff.js.map